@@ -43,6 +43,8 @@ from string import Template
 import yaml_header_tools
 
 import pandas as pd
+import logging
+
 
 import yaml
 
@@ -51,6 +53,12 @@ import yaml
 SPECIAL_PROPERTIES = ("description", "name", "id", "path",
                       "file", "checksum", "size")
 
+logger = logging.getLogger(__name__)
+
+
+class CrawlerTemplate(Template):
+    braceidpattern = r"(?a:[_a-z][_\.a-z0-9]*)"
+
 
 def _only_max(children_with_keys):
 
@@ -106,6 +114,19 @@ class ConverterValidationError(Exception):
         self.message = msg
 
 
+def create_path_value(func):
+    """decorator for create_values functions that adds a value containing the path
+
+    should be used for StructureElement that are associated with file system objects that have a
+    path, like File or Directory.
+    """
+
+    def inner(self, values: GeneralStore, element: StructureElement):
+        func(self, values=values, element=element)
+        values.update({self.name + ".path": element.path})
+    return inner
+
+
 def replace_variables(propvalue, values: GeneralStore):
     """
     This function replaces variables in property values (and possibly other locations,
@@ -129,7 +150,7 @@ def replace_variables(propvalue, values: GeneralStore):
             if isinstance(values[varname], db.Entity):
                 return values[varname]
 
-    propvalue_template = Template(propvalue)
+    propvalue_template = CrawlerTemplate(propvalue)
     return propvalue_template.safe_substitute(**values.get_storage())
 
 
@@ -237,7 +258,7 @@ def create_records(values: GeneralStore, records: RecordStore, def_records: dict
                 continue
 
             # Allow replacing variables in keys / names of properties:
-            key_template = Template(key)
+            key_template = CrawlerTemplate(key)
             key = key_template.safe_substitute(**values.get_storage())
 
             keys_modified.append((name, key))
@@ -325,9 +346,7 @@ class Converter(object, metaclass=ABCMeta):
 
         return converter
 
-    def create_values(self,
-                      values: GeneralStore,
-                      element: StructureElement):
+    def create_values(self, values: GeneralStore, element: StructureElement):
         """
         Extract information from the structure element and store them as values in the
         general store.
@@ -346,13 +365,14 @@ class Converter(object, metaclass=ABCMeta):
                         element: StructureElement):
         pass
 
-    def create_records(self, values: GeneralStore,
-                       records: RecordStore,
+    def create_records(self, values: GeneralStore, records: RecordStore,
                        element: StructureElement):
+        # TODO why is element passed but not used???
 
         if "records" not in self.definition:
             return []
 
+        # TODO please rename due to conflict
         return create_records(values,
                               records,
                               self.definition["records"])
@@ -364,7 +384,8 @@ class Converter(object, metaclass=ABCMeta):
 
         if rule not in FILTER_FUNCTIONS:
             raise RuntimeError(
-                f"{rule} is not a known filter rule. Only {list(FILTER_FUNCTIONS.keys())} are implemented."
+                f"{rule} is not a known filter rule. Only "
+                f"{list(FILTER_FUNCTIONS.keys())} are implemented."
             )
 
         to_be_filtered = []
@@ -391,19 +412,21 @@ class Converter(object, metaclass=ABCMeta):
         pass
 
     @staticmethod
-    def _debug_matching_template(name: str, regexp: list[str], matched: list[str], result: Optional[dict]):
+    def _debug_matching_template(name: str, regexp: list[str], matched: list[str],
+                                 result: Optional[dict]):
         """ Template for the debugging output for the match function """
-        print("\n--------", name, "-----------")
+        msg = "\n--------" + name + "-----------"
         for re, ma in zip(regexp, matched):
-            print("matching against:\n" + re)
-            print("matching:\n" + ma)
-            print("---------")
+            msg += "matching against:\n" + re
+            msg += "matching:\n" + ma
+            msg += "---------"
         if result is None:
-            print("No match")
+            msg += "No match"
         else:
-            print("Matched groups:")
-            print(result)
-        print("----------------------------------------")
+            msg += "Matched groups:"
+            msg += str(result)
+        msg += "----------------------------------------"
+        logger.debug(msg)
 
     @staticmethod
     def debug_matching(kind=None):
@@ -471,6 +494,10 @@ class DirectoryConverter(Converter):
 
         return children
 
+    @create_path_value
+    def create_values(self, values: GeneralStore, element: StructureElement):
+        super().create_values(values=values, element=element)
+
     def typecheck(self, element: StructureElement):
         return isinstance(element, Directory)
 
@@ -518,6 +545,10 @@ class SimpleFileConverter(Converter):
     def create_children(self, generalStore: GeneralStore, element: StructureElement):
         return list()
 
+    @create_path_value
+    def create_values(self, values: GeneralStore, element: StructureElement):
+        super().create_values(values=values, element=element)
+
     @Converter.debug_matching("name")
     def match(self, element: StructureElement):
         # TODO: See comment on types and inheritance
@@ -536,7 +567,7 @@ class FileConverter(SimpleFileConverter):
         super().__init__(*args, **kwargs)
 
 
-class MarkdownFileConverter(Converter):
+class MarkdownFileConverter(SimpleFileConverter):
     """
     reads the yaml header of markdown files (if a such a header exists).
     """
@@ -546,8 +577,18 @@ class MarkdownFileConverter(Converter):
         if not isinstance(element, File):
             raise RuntimeError("A markdown file is needed to create children.")
 
-        header = yaml_header_tools.get_header_from_file(
-            element.path, clean=False)
+        try:
+            header = yaml_header_tools.get_header_from_file(
+                element.path, clean=False)
+        except yaml_header_tools.NoValidHeader:
+            if generalStore is not None and self.name in generalStore:
+                path = generalStore[self.name]
+            else:
+                path = "<path not set>"
+            raise ConverterValidationError(
+                "Error during the validation (yaml header cannot be read) of the markdown file "
+                "located at the following node in the data structure:\n"
+                f"{path}")
         children: List[StructureElement] = []
 
         for name, entry in header.items():
@@ -560,25 +601,6 @@ class MarkdownFileConverter(Converter):
                     "Header entry {} has incompatible type.".format(name))
         return children
 
-    def typecheck(self, element: StructureElement):
-        return isinstance(element, File)
-
-    @Converter.debug_matching("name")
-    def match(self, element: StructureElement):
-        # TODO: See comment on types and inheritance
-        if not isinstance(element, File):
-            raise RuntimeError("Element must be a file.")
-        m = re.match(self.definition["match"], element.name)
-        if m is None:
-            return None
-        try:
-            yaml_header_tools.get_header_from_file(element.path)
-        except yaml_header_tools.NoValidHeader:
-            # TODO(salexan): Raise a validation error instead of just not
-            # matching silently.
-            return None
-        return m.groupdict()
-
 
 def convert_basic_element(element: Union[list, dict, bool, int, float, str, None], name=None,
                           msg_prefix=""):
@@ -685,20 +707,7 @@ class DictDictElementConverter(DictElementConverter):
         super().__init__(*args, **kwargs)
 
 
-class JSONFileConverter(Converter):
-    def typecheck(self, element: StructureElement):
-        return isinstance(element, File)
-
-    @Converter.debug_matching("name")
-    def match(self, element: StructureElement):
-        # TODO: See comment on types and inheritance
-        if not self.typecheck(element):
-            raise RuntimeError("Element must be a file")
-        m = re.match(self.definition["match"], element.name)
-        if m is None:
-            return None
-        return m.groupdict()
-
+class JSONFileConverter(SimpleFileConverter):
     def create_children(self, generalStore: GeneralStore, element: StructureElement):
         # TODO: See comment on types and inheritance
         if not isinstance(element, File):
@@ -714,26 +723,13 @@ class JSONFileConverter(Converter):
                     f"{element.path}\n" + err.message)
         structure_element = convert_basic_element(
             json_data,
-            name=element.name+"_child_dict",
+            name=element.name + "_child_dict",
             msg_prefix="The JSON File contained content that was parsed to a Python object"
             " with an unexpected type.")
         return [structure_element]
 
 
-class YAMLFileConverter(Converter):
-    def typecheck(self, element: StructureElement):
-        return isinstance(element, File)
-
-    @Converter.debug_matching("name")
-    def match(self, element: StructureElement):
-        # TODO: See comment on types and inheritance
-        if not self.typecheck(element):
-            raise RuntimeError("Element must be a file")
-        m = re.match(self.definition["match"], element.name)
-        if m is None:
-            return None
-        return m.groupdict()
-
+class YAMLFileConverter(SimpleFileConverter):
     def create_children(self, generalStore: GeneralStore, element: StructureElement):
         # TODO: See comment on types and inheritance
         if not isinstance(element, File):
@@ -749,7 +745,7 @@ class YAMLFileConverter(Converter):
                     f"{element.path}\n" + err.message)
         structure_element = convert_basic_element(
             yaml_data,
-            name=element.name+"_child_dict",
+            name=element.name + "_child_dict",
             msg_prefix="The YAML File contained content that was parsed to a Python object"
             " with an unexpected type.")
         return [structure_element]
@@ -1101,3 +1097,22 @@ class CSVTableConverter(TableConverter):
             child_elements.append(
                 DictElement(str(index), row.to_dict()))
         return child_elements
+
+
+class DateElementConverter(TextElementConverter):
+    """
+    allows to convert different text formats of dates to Python date objects.
+
+    The text to be parsed must be contained in the "date" group. The format string can be supplied
+    under "dateformat" in the Converter definition. The library used is datetime so see its
+    documentation for information on how to create the format string.
+    """
+
+    def match(self, element: StructureElement):
+        matches = super().match(element)
+        if matches is not None and "date" in matches:
+            matches.update({"date": datetime.datetime.strptime(
+                matches["date"],
+                self.definition["date_format"] if "date_format" in self.definition else "%Y-%m-%d"
+            ).date()})
+        return matches
