@@ -49,6 +49,7 @@ from typing import Any, Optional, Type, Union
 
 import caosdb as db
 
+from caosadvancedtools.utils import create_entity_link
 from caosadvancedtools.cache import UpdateCache, Cache
 from caosadvancedtools.crawler import Crawler as OldCrawler
 from caosdb.apiutils import (compare_entities, EntityMergeConflictError,
@@ -303,8 +304,7 @@ class Crawler(object):
                 # Validator is given by a path
                 if not value.startswith('/'):
                     # Not an absolute path
-                    definition[key] = os.path.join(
-                        os.path.dirname(definition_path), value)
+                    definition[key] = os.path.join(os.path.dirname(definition_path), value)
                     if not os.path.isfile(definition[key]):
                         # TODO(henrik) capture this in `crawler_main` similar to
                         # `ConverterValidationError`.
@@ -312,8 +312,7 @@ class Crawler(object):
                             f"Couldn't find validation file {definition[key]}")
             elif isinstance(value, dict):
                 # Recursively resolve all validators
-                definition[key] = self._resolve_validator_paths(
-                    value, definition_path)
+                definition[key] = self._resolve_validator_paths(value, definition_path)
 
         return definition
 
@@ -328,74 +327,13 @@ class Crawler(object):
         directory:
         schema.yml file
         README.md documentation
+
+        TODO: this function does not make use of self, so it could become static.
         """
 
         # Defaults for the converter registry:
-        converter_registry: dict[str, dict[str, str]] = {
-            "Directory": {
-                "converter": "DirectoryConverter",
-                "package": "caoscrawler.converters"},
-            "SimpleFile": {
-                "converter": "SimpleFileConverter",
-                "package": "caoscrawler.converters"},
-            "MarkdownFile": {
-                "converter": "MarkdownFileConverter",
-                "package": "caoscrawler.converters"},
-            "File": {
-                "converter": "SimpleFileConverter",
-                "package": "caoscrawler.converters"},
-            "JSONFile": {
-                "converter": "JSONFileConverter",
-                "package": "caoscrawler.converters"},
-            "YAMLFile": {
-                "converter": "YAMLFileConverter",
-                "package": "caoscrawler.converters"},
-            "CSVTableConverter": {
-                "converter": "CSVTableConverter",
-                "package": "caoscrawler.converters"},
-            "XLSXTableConverter": {
-                "converter": "XLSXTableConverter",
-                "package": "caoscrawler.converters"},
-            "DictBooleanElement": {
-                "converter": "BooleanElementConverter",
-                "package": "caoscrawler.converters"},
-            "BooleanElement": {
-                "converter": "BooleanElementConverter",
-                "package": "caoscrawler.converters"},
-            "DictFloatElement": {
-                "converter": "FloatElementConverter",
-                "package": "caoscrawler.converters"},
-            "FloatElement": {
-                "converter": "FloatElementConverter",
-                "package": "caoscrawler.converters"},
-            "DictTextElement": {
-                "converter": "TextElementConverter",
-                "package": "caoscrawler.converters"},
-            "TextElement": {
-                "converter": "TextElementConverter",
-                "package": "caoscrawler.converters"},
-            "DictIntegerElement": {
-                "converter": "IntegerElementConverter",
-                "package": "caoscrawler.converters"},
-            "IntegerElement": {
-                "converter": "IntegerElementConverter",
-                "package": "caoscrawler.converters"},
-            "DictListElement": {
-                "converter": "ListElementConverter",
-                "package": "caoscrawler.converters"},
-            "ListElement": {
-                "converter": "ListElementConverter",
-                "package": "caoscrawler.converters"},
-            "DictDictElement": {
-                "converter": "DictElementConverter",
-                "package": "caoscrawler.converters"},
-            "DictElement": {
-                "converter": "DictElementConverter",
-                "package": "caoscrawler.converters"},
-            "Dict": {
-                "converter": "DictElementConverter",
-                "package": "caoscrawler.converters"},
-        }
+        with open(str(files('caoscrawler').joinpath('default_converters.yml')), "r") as f:
+            converter_registry: dict[str, dict[str, str]] = yaml.safe_load(f)
 
         # More converters from definition file:
         if "Converters" in definition:
@@ -417,11 +355,16 @@ class Crawler(object):
             value["class"] = getattr(module, value["converter"])
         return converter_registry
 
-    def crawl_directory(self, dirname: str, crawler_definition_path: str):
+    def crawl_directory(self, dirname: str, crawler_definition_path: str,
+                        restricted_path: Optional[list[str]] = None):
         """ Crawl a single directory.
 
         Convenience function that starts the crawler (calls start_crawling)
         with a single directory as the StructureElement.
+
+        restricted_path: optional, list of strings
+                Traverse the data tree only along the given path. When the end of the given path
+                is reached, traverse the full tree as normal.
         """
 
         crawler_definition = self.load_definition(crawler_definition_path)
@@ -444,7 +387,9 @@ class Crawler(object):
         self.start_crawling(Directory(dir_structure_name,
                                       dirname),
                             crawler_definition,
-                            converter_registry)
+                            converter_registry,
+                            restricted_path=restricted_path
+                            )
 
     @staticmethod
     def initialize_converters(crawler_definition: dict, converter_registry: dict):
@@ -472,7 +417,8 @@ class Crawler(object):
 
     def start_crawling(self, items: Union[list[StructureElement], StructureElement],
                        crawler_definition: dict,
-                       converter_registry: dict):
+                       converter_registry: dict,
+                       restricted_path: Optional[list[str]] = None):
         """
         Start point of the crawler recursion.
 
@@ -484,6 +430,9 @@ class Crawler(object):
         crawler_definition : dict
              A dictionary representing the crawler definition, possibly from a yaml
              file.
+        restricted_path: optional, list of strings
+             Traverse the data tree only along the given path. When the end of the given path
+             is reached, traverse the full tree as normal.
 
         Returns
         -------
@@ -504,8 +453,14 @@ class Crawler(object):
 
         # This recursive crawling procedure generates the update list:
         self.crawled_data: list[db.Record] = []
-        self._crawl(items, local_converters, self.generalStore, self.recordStore, [], [])
-
+        self._crawl(
+            items=items,
+            local_converters=local_converters,
+            generalStore=self.generalStore,
+            recordStore=self.recordStore,
+            structure_elements_path=[],
+            converters_path=[],
+            restricted_path=restricted_path)
         if self.debug:
             self.debug_converters = local_converters
 
@@ -942,16 +897,17 @@ class Crawler(object):
         because some changes in parents (e.g. of Files) might fail
         if they are not updated first.
         """
+        logger.debug("=== Going to execute parent updates ===")
         Crawler.set_ids_and_datatype_of_parents_and_properties(to_be_updated)
         parent_updates = db.Container()
 
-        for record in to_be_updated:
-            old_entity = Crawler._get_entity_by_id(record.id)
+        for entity in to_be_updated:
+            old_entity = Crawler._get_entity_by_id(entity.id)
 
             # Check whether the parents have been changed and add them if missing
             # in the old entity:
             changes_made = False
-            for parent in record.parents:
+            for parent in entity.parents:
                 found = False
                 for old_parent in old_entity.parents:
                     if old_parent.id == parent.id:
@@ -1061,20 +1017,25 @@ class Crawler(object):
                                                                            referencing_entities)
             for record in to_be_updated]
         # Merge with existing data to prevent unwanted overwrites
-        to_be_updated = self._merge_properties_from_remote(to_be_updated,
-                                                           identified_records)
+        to_be_updated = self._merge_properties_from_remote(to_be_updated, identified_records)
         # remove unnecessary updates from list by comparing the target records
         # to the existing ones
-        to_be_updated = self.remove_unnecessary_updates(
-            to_be_updated, identified_records)
+        to_be_updated = self.remove_unnecessary_updates(to_be_updated, identified_records)
 
+        logger.info(f"Going to insert {len(to_be_inserted)} Entities and update "
+                    f"{len(to_be_inserted)} Entities.")
         if commit_changes:
             self.execute_parent_updates_in_list(to_be_updated, securityMode=self.securityMode,
                                                 run_id=self.run_id, unique_names=unique_names)
+            logger.info(f"Added parent RecordTypes where necessary.")
             self.execute_inserts_in_list(
                 to_be_inserted, self.securityMode, self.run_id, unique_names=unique_names)
+            logger.info(f"Executed inserts:\n"
+                        + self.create_entity_summary(to_be_inserted))
             self.execute_updates_in_list(
                 to_be_updated, self.securityMode, self.run_id, unique_names=unique_names)
+            logger.info(f"Executed updates:\n"
+                        + self.create_entity_summary(to_be_updated))
 
         update_cache = UpdateCache()
         pending_inserts = update_cache.get_inserts(self.run_id)
@@ -1089,6 +1050,25 @@ class Crawler(object):
 
         return (to_be_inserted, to_be_updated)
 
+    @staticmethod
+    def create_entity_summary(entities: list[db.Entity]):
+        """ Creates a summary string reprensentation of a list of entities."""
+        parents = {}
+        for el in entities:
+            for pp in el.parents:
+                if pp.name not in parents:
+                    parents[pp.name] = [el]
+                else:
+                    parents[pp.name].append(el)
+        output = ""
+        for key, value in parents.items():
+            output += f"{key}:\n"
+            for el in value:
+                output += create_entity_link(el) + ", "
+
+            output = output[:-2] + "\n"
+        return output
+
     @staticmethod
     def inform_about_pending_changes(pending_changes, run_id, path, inserts=False):
         # Sending an Email with a link to a form to authorize updates is
@@ -1156,11 +1136,14 @@ ____________________\n""".format(i + 1, len(pending_changes)) + str(el[3]))
         with open(filename, "w") as f:
             f.write(yaml.dump(paths, sort_keys=False))
 
-    def _crawl(self, items: list[StructureElement],
+    def _crawl(self,
+               items: list[StructureElement],
                local_converters: list[Converter],
                generalStore: GeneralStore,
                recordStore: RecordStore,
-               structure_elements_path: list[str], converters_path: list[str]):
+               structure_elements_path: list[str],
+               converters_path: list[str],
+               restricted_path: Optional[list[str]] = None):
         """
         Crawl a list of StructureElements and apply any matching converters.
 
@@ -1169,20 +1152,35 @@ ____________________\n""".format(i + 1, len(pending_changes)) + str(el[3]))
                             treating structure elements. A locally defined converter could be
                             one that is only valid for a specific subtree of the originally
                             cralwed StructureElement structure.
-        generalStore and recordStore: This recursion of the crawl function should only operate on copies of the
-                            global stores of the Crawler object.
+        generalStore and recordStore: This recursion of the crawl function should only operate on
+                                      copies of the global stores of the Crawler object.
+        restricted_path: optional, list of strings, traverse the data tree only along the given
+                         path. For example, when a directory contains files a, b and c and b is
+                         given in restricted_path, a and c will be ignroed by the crawler.
+                         When the end of the given path is reached, traverse the full tree as
+                         normal. The first element of the list provided by restricted_path should
+                         be the name of the StructureElement at this level, i.e. denoting the
+                         respective element in the items argument.
         """
+        # This path_found variable stores wether the path given by restricted_path was found in the
+        # data tree
+        path_found = False
+        if restricted_path is not None and len(restricted_path) == 0:
+            restricted_path = None
+
         for element in items:
             for converter in local_converters:
 
                 # type is something like "matches files", replace isinstance with "type_matches"
                 # match function tests regexp for example
-                if (converter.typecheck(element) and
-                        converter.match(element) is not None):
+                if (converter.typecheck(element) and (
+                        restricted_path is None or element.name == restricted_path[0])
+                        and converter.match(element) is not None):
+                    path_found = True
                     generalStore_copy = generalStore.create_scoped_copy()
                     recordStore_copy = recordStore.create_scoped_copy()
 
-                    # Create an entry for this matched structure element:
+                    # Create an entry for this matched structure element that contains the path:
                     generalStore_copy[converter.name] = (
                         os.path.join(*(structure_elements_path + [element.get_name()])))
 
@@ -1196,7 +1194,7 @@ ____________________\n""".format(i + 1, len(pending_changes)) + str(el[3]))
                     children = converter.create_children(generalStore_copy, element)
 
                     if self.debug:
-                        # add provenance information for each varaible
+                        # add provenance information for each variable
                         self.debug_tree[str(element)] = (
                             generalStore_copy.get_storage(), recordStore_copy.get_storage())
                         self.debug_metadata["copied"][str(element)] = (
@@ -1219,7 +1217,12 @@ ____________________\n""".format(i + 1, len(pending_changes)) + str(el[3]))
                     self._crawl(children, converter.converters,
                                 generalStore_copy, recordStore_copy,
                                 structure_elements_path + [element.get_name()],
-                                converters_path + [converter.name])
+                                converters_path + [converter.name],
+                                restricted_path[1:] if restricted_path is not None else None)
+
+        if restricted_path and not path_found:
+            raise RuntimeError("A 'restricted_path' argument was given that is not contained in "
+                               "the data tree")
         # if the crawler is running out of scope, copy all records in
         # the recordStore, that were created in this scope
         # to the general update container.
@@ -1250,6 +1253,9 @@ def crawler_main(crawled_directory_path: str,
                  prefix: str = "",
                  securityMode: SecurityMode = SecurityMode.UPDATE,
                  unique_names=True,
+                 restricted_path: Optional[list[str]] = None,
+                 remove_prefix: Optional[str] = None,
+                 add_prefix: Optional[str] = None,
                  ):
     """
 
@@ -1268,11 +1274,18 @@ def crawler_main(crawled_directory_path: str,
     dry_run : bool
         do not commit any chnages to the server
     prefix : str
-        remove the given prefix from file paths
+        DEPRECATED, remove the given prefix from file paths
     securityMode : int
         securityMode of Crawler
     unique_names : bool
         whether or not to update or insert entities inspite of name conflicts
+    restricted_path: optional, list of strings
+            Traverse the data tree only along the given path. When the end of the given path
+            is reached, traverse the full tree as normal.
+    remove_prefix : Optional[str]
+        remove the given prefix from file paths
+    add_prefix : Optional[str]
+        add the given prefix to file paths
 
     Returns
     -------
@@ -1281,19 +1294,27 @@ def crawler_main(crawled_directory_path: str,
     """
     crawler = Crawler(debug=debug, securityMode=securityMode)
     try:
-        crawler.crawl_directory(crawled_directory_path, cfood_file_name)
+        crawler.crawl_directory(crawled_directory_path, cfood_file_name, restricted_path)
     except ConverterValidationError as err:
-        print(err)
+        logger.error(err)
         return 1
-    if provenance_file is not None:
+    if provenance_file is not None and debug:
         crawler.save_debug_data(provenance_file)
 
     if identifiables_definition_file is not None:
-
         ident = CaosDBIdentifiableAdapter()
         ident.load_from_yaml_definition(identifiables_definition_file)
         crawler.identifiableAdapter = ident
 
+    if prefix != "":
+        warnings.warn(DeprecationWarning("The prefix argument is deprecated and will be removed "
+                                         "in the future. Please use `remove_prefix` instead."))
+        if remove_prefix is not None:
+            raise ValueError("Please do not supply the (deprecated) `prefix` and the "
+                             "`remove_prefix` argument at the same time. Only use "
+                             "`remove_prefix` instead.")
+        remove_prefix = prefix
+
     if dry_run:
         ins, upd = crawler.synchronize(commit_changes=False)
         inserts = [str(i) for i in ins]
@@ -1308,11 +1329,15 @@ def crawler_main(crawled_directory_path: str,
             if isinstance(elem, db.File):
                 # correct the file path:
                 # elem.file = os.path.join(args.path, elem.file)
-                if prefix is None:
-                    raise RuntimeError(
-                        "No prefix set. Prefix must be set if files are used.")
-                if elem.path.startswith(prefix):
-                    elem.path = elem.path[len(prefix):]
+                if remove_prefix:
+                    if elem.path.startswith(remove_prefix):
+                        elem.path = elem.path[len(remove_prefix):]
+                    else:
+                        raise RuntimeError("Prefix shall be removed from file path but the path "
+                                           "does not start with the prefix:"
+                                           f"\n{remove_prefix}\n{elem.path}")
+                if add_prefix:
+                    elem.path = add_prefix + elem.path
                 elem.file = None
                 # TODO: as long as the new file backend is not finished
                 #       we are using the loadFiles function to insert symlinks.
@@ -1346,6 +1371,15 @@ def parse_args():
                                      formatter_class=RawTextHelpFormatter)
     parser.add_argument("cfood_file_name",
                         help="Path name of the cfood yaml file to be used.")
+    mg = parser.add_mutually_exclusive_group()
+    mg.add_argument("-r", "--restrict", nargs="*",
+                    help="Restrict the crawling to the subtree at the end of the given path."
+                    "I.e. for each level that is given the crawler only treats the element "
+                    "with the given name.")
+    mg.add_argument("--restrict-path", help="same as restrict; instead of a list, this takes a "
+                    "single string that is interpreded as file system path. Note that a trailing"
+                    "separator (e.g. '/') will be ignored. Use --restrict if you need to have "
+                    "empty strings.")
     parser.add_argument("--provenance", required=False,
                         help="Path name of the provenance yaml file. "
                         "This file will only be generated if this option is set.")
@@ -1371,18 +1405,35 @@ def parse_args():
     parser.add_argument("-u", "--unique-names",
                         help="Insert or updates entities even if name conflicts exist.")
     parser.add_argument("-p", "--prefix",
-                        help="Remove the given prefix from the paths "
-                        "of all file objects.")
+                        help="DEPRECATED, use --remove-prefix instead. Remove the given prefix "
+                        "from the paths of all file objects.")
+    parser.add_argument("--remove-prefix",
+                        help="Remove the given prefix from the paths of all file objects.")
+    parser.add_argument("--add-prefix",
+                        help="Add the given prefix to the paths of all file objects.")
 
     return parser.parse_args()
 
 
+def split_restricted_path(path):
+    elements = []
+    while path != "/":
+        path, el = os.path.split(path)
+        if el != "":
+            elements.insert(0, el)
+    return elements
+
+
 def main():
     args = parse_args()
 
     conlogger = logging.getLogger("connection")
     conlogger.setLevel(level=logging.ERROR)
 
+    if args.prefix:
+        print("Please use '--remove-prefix' option instead of '--prefix' or '-p'.")
+        return -1
+
     # logging config for local execution
     logger.addHandler(logging.StreamHandler(sys.stdout))
     if args.debug:
@@ -1392,6 +1443,12 @@ def main():
 
     if args.add_cwd_to_path:
         sys.path.append(os.path.abspath("."))
+    restricted_path = None
+    if args.restrict_path:
+        restricted_path = split_restricted_path(args.restrict_path)
+    if args.restrict:
+        restricted_path = args.restrict
+
     sys.exit(crawler_main(
         crawled_directory_path=args.crawled_directory_path,
         cfood_file_name=args.cfood_file_name,
@@ -1399,11 +1456,13 @@ def main():
         debug=args.debug,
         provenance_file=args.provenance,
         dry_run=args.dry_run,
-        prefix=args.prefix,
         securityMode={"retrieve": SecurityMode.RETRIEVE,
                       "insert": SecurityMode.INSERT,
                       "update": SecurityMode.UPDATE}[args.security_mode],
         unique_names=args.unique_names,
+        restricted_path=restricted_path,
+        remove_prefix=args.remove_prefix,
+        add_prefix=args.add_prefix,
     ))
 
 
