@@ -33,6 +33,7 @@ import caosdb as db
 import logging
 from abc import abstractmethod, ABCMeta
 from .utils import has_parent
+
 logger = logging.getLogger(__name__)
 
 
@@ -447,7 +448,7 @@ class CaosDBIdentifiableAdapter(IdentifiableAdapter):
     def get_file(self, identifiable: Identifiable):
         if identifiable.path is None:
             raise RuntimeError("Path must not be None for File retrieval.")
-        candidates = db.execute_query("FIND File which is stored at {}".format(
+        candidates = db.execute_query("FIND File which is stored at '{}'".format(
             identifiable.path))
         if len(candidates) > 1:
             raise RuntimeError("Identifiable was not defined unambigiously.")
